import re

import mongomock
import pymongo

import superduperdb as s
from superduperdb.backends.base.backends import data_backends, metadata_stores
from superduperdb.backends.filesystem.artifacts import FileSystemArtifactStore
from superduperdb.backends.mongodb.artifacts import MongoArtifactStore
from superduperdb.base.datalayer import Datalayer
from superduperdb.server.dask_client import dask_client


def build_artifact_store(cfg):
    if cfg.artifact_store is None:
        raise ValueError('No artifact store specified')
    elif cfg.artifact_store.startswith('mongodb://'):
        conn = pymongo.MongoClient('/'.join(cfg.artifact_store.split('/')[:-1]))
        name = cfg.artifact_store.split('/')[-1]
        return MongoArtifactStore(conn, name)
    elif cfg.artifact_store.startswith('filesystem://'):
        directory = cfg.artifact_store.split('://')[1]
        return FileSystemArtifactStore(directory)
    else:
        raise ValueError(f'Unknown artifact store: {cfg.artifact_store}')


def build_datalayer(cfg=None, **kwargs) -> Datalayer:
    """
    Build db as per ``db = superduper(db)`` from configuration.

    :param cfg: configuration to use. If None, use ``superduperdb.CFG``.
    """
    cfg = cfg or s.CFG

    for k, v in kwargs.items():
        setattr(cfg, k, v)

    def build(uri, mapping):
        if re.match('^mongodb:\/\/|^mongodb\+srv:\/\/', uri) is not None:
            name = uri.split('/')[-1]
            uri = '/'.join(uri.split('/')[:-1])
            conn = pymongo.MongoClient(uri, serverSelectionTimeoutMS=5000)
            return mapping['mongodb'](conn, name)
        elif uri.startswith('mongomock://'):
            name = uri.split('/')[-1]
            conn = mongomock.MongoClient()
            return mapping['mongodb'](conn, name)
        else:
            import ibis

            conn = ibis.connect(uri)
            return mapping['ibis'](conn, uri.split('//')[0])

    databackend = build(cfg.data_backend, data_backends)

    db = Datalayer(
        databackend=databackend,
        metadata=(
            build(cfg.metadata_store, metadata_stores)
            if cfg.metadata_store is not None
            else databackend.build_metadata()
        ),
        artifact_store=(
            build_artifact_store(cfg)
            if cfg.artifact_store is not None
            else databackend.build_artifact_store()
        ),
        distributed_client=dask_client(
            cfg.cluster.dask_scheduler,
            local=cfg.cluster.local,
            serializers=cfg.cluster.serializers,
            deserializers=cfg.cluster.deserializers,
        )
        if cfg.cluster.distributed
        else None,
    )

    return db



import typing as t

import superduperdb as s

from . import command


@command(help='Start server')
def serve():
    from superduperdb.base.build import build_datalayer
    from superduperdb.server.server import serve

    db = build_datalayer()
    serve(db)


@command(help='Start local cluster: server, dask and change data capture')
def local_cluster(on: t.List[str] = []):
    from superduperdb.backends.mongodb.query import Collection
    from superduperdb.base.build import build_datalayer
    from superduperdb.server.dask_client import dask_client
    from superduperdb.server.server import serve

    db = build_datalayer()
    dask_client(
        uri=s.CFG.cluster.dask_scheduler,
        local=True,
    )
    for collection in on:
        db.cdc.listen(
            on=Collection(identifier=collection),
        )
    serve(db)



import dataclasses as dc
import typing as t

from superduperdb.components.component import Component
from superduperdb.misc.serialization import serializers

if t.TYPE_CHECKING:
    from superduperdb.base.datalayer import Datalayer


@dc.dataclass
class Serializer(Component):
    identifier: str
    object: t.Type

    type_id: t.ClassVar[str] = 'serializer'

    version: t.Optional[int]

    def on_create(self, db: 'Datalayer'):
        serializers.add(self.identifier, self.object)

        self.object = t.cast(t.Type, self.identifier)



import dataclasses as dc

from superduperdb.base.cursor import SuperDuperCursor
from superduperdb.base.document import Document


class IbisDocument(Document):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def outputs(self, *args):
        return self.__getitem__('output')


@dc.dataclass
class SuperDuperIbisCursor(SuperDuperCursor):
    def wrap_document(self, r, encoders):
        """
        Wrap a document in a ``Document``.
        """
        return IbisDocument(Document.decode(r, encoders))



from __future__ import annotations

import dataclasses as dc
import logging
import multiprocessing
import typing as t
from functools import wraps

import tqdm
from dask.distributed import Future
from numpy import ndarray
from sklearn.pipeline import Pipeline

import superduperdb as s
from superduperdb.backends.base.query import CompoundSelect, Select
from superduperdb.backends.ibis.field_types import FieldType
from superduperdb.backends.query_dataset import QueryDataset
from superduperdb.base.artifact import Artifact
from superduperdb.base.serializable import Serializable
from superduperdb.components.component import Component
from superduperdb.components.dataset import Dataset
from superduperdb.components.encoder import Encoder
from superduperdb.components.metric import Metric
from superduperdb.components.schema import Schema
from superduperdb.jobs.job import ComponentJob, Job
from superduperdb.misc.special_dicts import MongoStyleDict

if t.TYPE_CHECKING:
    from superduperdb.base.datalayer import Datalayer

EncoderArg = t.Union[Encoder, FieldType, str, None]
ObjectsArg = t.Sequence[t.Union[t.Any, Artifact]]


@dc.dataclass
class _TrainingConfiguration(Component):
    """
    Training configuration object, containing all settings necessary for a particular
    learning-task use-case to be serialized and initiated. The object is ``callable``
    and returns a class which may be invoked to apply training.

    :param identifier: Unique identifier of configuration
    :param **kwargs: Key-values pairs, the variables which configure training.
    :param version: Version number of the configuration
    """

    identifier: str
    kwargs: t.Optional[t.Dict] = None
    version: t.Optional[int] = None

    type_id: t.ClassVar[str] = 'training_configuration'

    def get(self, k, default=None):
        try:
            return getattr(self, k)
        except AttributeError:
            return self.kwargs.get(k, default=default)


class PredictMixin:
    """
    Mixin class for components which can predict.

    :param identifier: Unique identifier of model
    :param encoder: Encoder instance (optional)
    :param preprocess: Preprocess function (optional)
    :param postprocess: Postprocess function (optional)
    :param collate_fn: Collate function (optional)
    :param batch_predict: Whether to batch predict (optional)
    :param takes_context: Whether the model takes context into account (optional)
    :param to_call: The method to use for prediction (optional)
    """

    identifier: str
    encoder: EncoderArg
    output_schema: t.Optional[t.Union[Schema, dict]] = None
    flatten: bool = False
    preprocess: t.Union[t.Callable, Artifact, None] = None
    postprocess: t.Union[t.Callable, Artifact, None] = None
    collate_fn: t.Union[t.Callable, Artifact, None] = None
    batch_predict: bool
    takes_context: bool
    to_call: t.Callable
    model_update_kwargs: t.Dict

    def create_predict_job(
        self,
        X: str,
        select: t.Optional[Select] = None,
        ids: t.Optional[t.Sequence[str]] = None,
        max_chunk_size: t.Optional[int] = None,
        **kwargs,
    ):
        return ComponentJob(
            component_identifier=self.identifier,
            method_name='predict',
            type_id='model',
            args=[X],
            kwargs={
                'distributed': False,
                'select': select.serialize() if select else None,
                'ids': ids,
                'max_chunk_size': max_chunk_size,
                **kwargs,
            },
        )

    async def _apredict_one(self, X: t.Any, **kwargs):
        raise NotImplementedError

    async def _apredict(self, X: t.Any, one: bool = False, **kwargs):
        raise NotImplementedError

    def _predict_one(self, X: t.Any, **kwargs) -> int:
        if isinstance(self.preprocess, Artifact):
            X = self.preprocess.artifact(X)
        elif self.preprocess is not None:
            raise ValueError('Bad preprocess')

        output = self.to_call(X, **kwargs)

        if isinstance(self.postprocess, Artifact):
            output = self.postprocess.artifact(output)
        elif self.postprocess is not None:
            raise ValueError('Bad postprocess')

        return output

    def _forward(self, X: t.Sequence[int], num_workers: int = 0) -> t.Sequence[int]:
        if self.batch_predict:
            return self.to_call(X)

        outputs = []
        if num_workers:
            pool = multiprocessing.Pool(processes=num_workers)
            for r in pool.map(self.to_call, X):
                outputs.append(r)
            pool.close()
            pool.join()
        else:
            for r in X:
                outputs.append(self.to_call(r))
        return outputs

    def _predict(
        self, X: t.Any, one: bool = False, **predict_kwargs
    ) -> t.Union[ndarray, int, t.Sequence[int]]:
        if one:
            return self._predict_one(X)

        if isinstance(self.preprocess, Artifact):
            X = [self.preprocess.artifact(i) for i in X]
        elif self.preprocess is not None:
            raise ValueError('Bad preprocess')

        if isinstance(self.collate_fn, Artifact):
            raise ValueError('Bad collate function')
        elif self.collate_fn is not None:
            X = self.collate_fn(X)

        outputs = self._forward(X, **predict_kwargs)

        if isinstance(self.postprocess, Artifact):
            outputs = [self.postprocess.artifact(o) for o in outputs]
        elif self.postprocess is not None:
            raise ValueError('Bad postprocess')

        return outputs

    def predict(
        self,
        X: t.Any,
        db: t.Optional[Datalayer] = None,
        select: t.Optional[Select] = None,
        distributed: t.Optional[bool] = None,
        ids: t.Optional[t.List[str]] = None,
        max_chunk_size: t.Optional[int] = None,
        dependencies: t.Sequence[Job] = (),
        listen: bool = False,
        one: bool = False,
        context: t.Optional[t.Dict] = None,
        in_memory: bool = True,
        overwrite: bool = False,
        **kwargs,
    ) -> t.Any:
        if one:
            assert db is None, 'db must be None when ``one=True`` (direct call)'

        if isinstance(select, dict):
            select = Serializable.deserialize(select)

        if db is not None:
            logging.info(f'Adding model {self.identifier} to db')
            assert isinstance(self, Component)
            db.add(self)
            logging.info('Done.')

        if distributed is None:
            distributed = s.CFG.cluster.distributed

        if listen:
            assert db is not None
            assert select is not None
            return self._predict_and_listen(
                X=X, db=db, select=select, max_chunk_size=max_chunk_size, **kwargs
            )

        if distributed:
            return self.create_predict_job(
                X,
                select=select,
                ids=ids,
                max_chunk_size=max_chunk_size,
                overwrite=overwrite,
                **kwargs,
            )(db=db, distributed=distributed, dependencies=dependencies)
        else:
            if select is not None and ids is None:
                assert db is not None
                return self._predict_with_select(
                    X=X,
                    select=select,
                    db=db,
                    in_memory=in_memory,
                    max_chunk_size=max_chunk_size,
                    overwrite=overwrite,
                    **kwargs,
                )
            elif select is not None and ids is not None:
                assert db is not None
                return self._predict_with_select_and_ids(
                    X=X,
                    select=select,
                    ids=ids,
                    db=db,
                    max_chunk_size=max_chunk_size,
                    in_memory=in_memory,
                    **kwargs,
                )
            else:
                if self.takes_context:
                    kwargs['context'] = context
                return self._predict(X, one=one, **kwargs)

    async def apredict(
        self,
        X: t.Any,
        context: t.Optional[t.Dict] = None,
        one: bool = False,
        **kwargs,
    ):
        if self.takes_context:
            kwargs['context'] = context
        return await self._apredict(X, one=one, **kwargs)

    def _predict_and_listen(
        self,
        X: t.Any,
        select: CompoundSelect,
        db: Datalayer,
        in_memory: bool = True,
        max_chunk_size: t.Optional[int] = None,
        dependencies: t.Sequence[Job] = (),
        **kwargs,
    ):
        from superduperdb.components.listener import Listener

        return db.add(
            Listener(
                key=X,
                model=t.cast(Model, self),
                select=select,
                predict_kwargs={
                    **kwargs,
                    'in_memory': in_memory,
                    'max_chunk_size': max_chunk_size,
                },
            ),
            dependencies=dependencies,
        )

    def _predict_with_select(
        self,
        X: t.Any,
        select: Select,
        db: Datalayer,
        max_chunk_size: t.Optional[int] = None,
        in_memory: bool = True,
        overwrite: bool = False,
        **kwargs,
    ):
        ids = []
        if overwrite:
            query = select.select_ids
        else:
            query = select.select_ids_of_missing_outputs(key=X, model=self.identifier)

        for r in tqdm.tqdm(db.execute(query)):
            ids.append(str(r[db.databackend.id_field]))

        return self._predict_with_select_and_ids(
            X=X,
            db=db,
            ids=ids,
            select=select,
            max_chunk_size=max_chunk_size,
            in_memory=in_memory,
            **kwargs,
        )

    def _predict_with_select_and_ids(
        self,
        X: t.Any,
        db: Datalayer,
        select: Select,
        ids: t.List[str],
        in_memory: bool = True,
        max_chunk_size: t.Optional[int] = None,
        **kwargs,
    ):
        if max_chunk_size is not None:
            it = 0
            for i in range(0, len(ids), max_chunk_size):
                logging.info(f'Computing chunk {it}/{int(len(ids) / max_chunk_size)}')
                self._predict_with_select_and_ids(
                    X=X,
                    db=db,
                    ids=ids[i : i + max_chunk_size],
                    select=select,
                    max_chunk_size=None,
                    in_memory=in_memory,
                    **kwargs,
                )
                it += 1
            return

        X_data: t.Any
        if in_memory:
            if db is None:
                raise ValueError('db cannot be None')
            docs = list(db.execute(select.select_using_ids(ids)))
            if X != '_base':
                X_data = [MongoStyleDict(r.unpack())[X] for r in docs]
            else:
                X_data = [r.unpack() for r in docs]
        else:
            X_data = QueryDataset(
                select=select,
                ids=ids,
                fold=None,
                db=db,
                in_memory=False,
                keys=[X],
            )

        outputs = self.predict(X=X_data, one=False, distributed=False, **kwargs)

        if isinstance(self.encoder, Encoder):
            if self.flatten:
                outputs = [
                    [self.encoder(x).encode() for x in output] for output in outputs
                ]
            else:
                outputs = [self.encoder(x).encode() for x in outputs]
        elif isinstance(self.output_schema, Schema):
            encoded_ouputs = []
            for output in outputs:
                if isinstance(output, dict):
                    encoded_ouputs.append(self.output_schema.encode(output))
                elif self.flatten:
                    encoded_output = [self.output_schema.encode(x) for x in output]
                    encoded_ouputs.append(encoded_output)
            outputs = encoded_ouputs if encoded_ouputs else outputs

        select.model_update(
            db=db,
            model=self.identifier,
            outputs=outputs,
            key=X,
            ids=ids,
            flatten=self.flatten,
            **self.model_update_kwargs,
        )
        return


@dc.dataclass
class Model(Component, PredictMixin):
    """Model component which wraps a model to become serializable

    :param identifier: Unique identifier of model
    :param object: Model object, e.g. sklearn model, etc..
    :param encoder: Encoder instance (optional)
    :param flatten: Flatten the model outputs
    :param output_schema: Output schema (mapping of encoders) (optional)
    :param preprocess: Preprocess function (optional)
    :param postprocess: Postprocess function (optional)
    :param collate_fn: Collate function (optional)
    :param metrics: Metrics to use (optional)
    :param predict_method: The method to use for prediction (optional)
    :param model_to_device_method: The method to transfer the model to a device
    :param batch_predict: Whether to batch predict (optional)
    :param takes_context: Whether the model takes context into account (optional)
    :param serializer: Serializer to store model to artifact store(optional)
    """

    identifier: str
    object: t.Union[Artifact, t.Any]
    flatten: bool = False
    output_schema: t.Optional[t.Union[Schema, dict]] = None
    encoder: EncoderArg = None
    preprocess: t.Union[t.Callable, Artifact, None] = None
    postprocess: t.Union[t.Callable, Artifact, None] = None
    collate_fn: t.Union[t.Callable, Artifact, None] = None
    metrics: t.Sequence[t.Union[str, Metric, None]] = ()
    predict_method: t.Optional[str] = None
    model_to_device_method: t.Optional[str] = None
    batch_predict: bool = False
    takes_context: bool = False
    train_X: t.Optional[str] = None
    train_y: t.Optional[str] = None
    training_select: t.Union[Select, None] = None
    metric_values: t.Optional[t.Dict] = dc.field(default_factory=dict)
    training_configuration: t.Union[str, _TrainingConfiguration, None] = None
    model_update_kwargs: dict = dc.field(default_factory=dict)
    serializer: str = 'dill'

    version: t.Optional[int] = None
    future: t.Optional[Future] = None
    device: str = "cpu"

    # TODO: handle situation with multiple GPUs
    preferred_devices: t.Union[None, t.Sequence[str]] = ("cuda", "mps", "cpu")

    artifact_attributes: t.ClassVar[t.Sequence[str]] = ['object']

    type_id: t.ClassVar[str] = 'model'

    def __post_init__(self):
        if not isinstance(self.object, Artifact):
            self.object = Artifact(artifact=self.object, serializer=self.serializer)
        if self.preprocess and not isinstance(self.preprocess, Artifact):
            self.preprocess = Artifact(artifact=self.preprocess)
        if self.postprocess and not isinstance(self.postprocess, Artifact):
            self.postprocess = Artifact(artifact=self.postprocess)
        if self.predict_method is None:
            self.to_call = self.object.artifact
        else:
            self.to_call = getattr(self.object.artifact, self.predict_method)

        self._artifact_method = None

        if self.model_to_device_method is not None:
            self._artifact_method = getattr(self, self.model_to_device_method)

    def on_load(self, db: Datalayer) -> None:
        if self._artifact_method and self.preferred_devices:
            for i, device in enumerate(self.preferred_devices):
                try:
                    self._artifact_method(device)
                    self.device = device
                    return
                except Exception:
                    if i == len(self.preferred_devices) - 1:
                        raise
        if isinstance(self.output_schema, Schema):
            db.add(self.output_schema)

    @property
    def child_components(self) -> t.Sequence[t.Tuple[str, str]]:
        out = []
        if isinstance(self.encoder, Encoder):
            out.append(('encoder', 'encoder'))
        if self.training_configuration is not None:
            out.append(('training_configuration', 'training_configuration'))
        return out

    @property
    def training_keys(self) -> t.List:
        if isinstance(self.train_X, list):
            out = list(self.train_X)
        elif self.train_X is not None:
            out = [self.train_X]
        if self.train_y is not None:
            if isinstance(self.train_y, list):
                out.extend(self.train_y)
            else:
                out.append(self.train_y)
        return out

    def append_metrics(self, d: t.Dict[str, float]) -> None:
        if self.metric_values is not None:
            for k, v in d.items():
                self.metric_values.setdefault(k, []).append(v)

    def validate(
        self, db, validation_set: t.Union[Dataset, str], metrics: t.Sequence[Metric]
    ):
        db.add(self)
        out = self._validate(db, validation_set, metrics)
        if self.metric_values is None:
            raise ValueError('self.metric_values cannot be None')
        self.metric_values.update(out)
        db.metadata.update_object(
            type_id='model',
            identifier=self.identifier,
            version=self.version,
            key='dict.metric_values',
            value=self.metric_values,
        )

    def on_create(self, db: Datalayer):
        if isinstance(self.encoder, str):
            self.encoder = db.load('encoder', self.encoder)  # type: ignore[assignment]
        # TODO: check if output table should be created
        output_component = db.databackend.create_model_table_or_collection(self)
        if output_component is not None:
            db.add(output_component)

    def _validate(
        self,
        db: Datalayer,
        validation_set: t.Union[Dataset, str],
        metrics: t.Sequence[Metric],
    ):
        if isinstance(validation_set, str):
            validation_set = t.cast(Dataset, db.load('dataset', validation_set))

        mdicts = [MongoStyleDict(r.unpack()) for r in validation_set.data]
        assert self.train_X is not None
        prediction = self._predict([d[self.train_X] for d in mdicts])
        assert self.train_y is not None
        target = [d[self.train_y] for d in mdicts]
        assert isinstance(prediction, list)
        assert isinstance(target, list)
        results = {}

        for m in metrics:
            out = m(prediction, target)
            results[f'{validation_set.identifier}/{m.identifier}'] = out
        return results

    def create_fit_job(
        self,
        X: t.Union[str, t.Sequence[str]],
        select: t.Optional[Select] = None,
        y: t.Optional[str] = None,
        **kwargs,
    ):
        return ComponentJob(
            component_identifier=self.identifier,
            method_name='fit',
            type_id='model',
            args=[X],
            kwargs={
                'y': y,
                'distributed': False,
                'select': select.serialize() if select else None,
                **kwargs,
            },
        )

    def _fit(
        self,
        X: t.Any,
        y: t.Any = None,
        configuration: t.Optional[_TrainingConfiguration] = None,
        data_prefetch: bool = False,
        db: t.Optional[Datalayer] = None,
        dependencies: t.Sequence[Job] = (),
        metrics: t.Optional[t.Sequence[Metric]] = None,
        select: t.Optional[Select] = None,
        validation_sets: t.Optional[t.Sequence[t.Union[str, Dataset]]] = None,
    ):
        raise NotImplementedError

    def fit(
        self,
        X: t.Any,
        y: t.Any = None,
        configuration: t.Optional[_TrainingConfiguration] = None,
        data_prefetch: bool = False,
        db: t.Optional[Datalayer] = None,
        dependencies: t.Sequence[Job] = (),
        distributed: t.Optional[bool] = None,
        metrics: t.Optional[t.Sequence[Metric]] = None,
        select: t.Optional[Select] = None,
        validation_sets: t.Optional[t.Sequence[t.Union[str, Dataset]]] = None,
        **kwargs,
    ) -> t.Optional[Pipeline]:
        if isinstance(select, dict):
            select = Serializable.deserialize(select)

        if validation_sets:
            validation_sets = list(validation_sets)
            for i, vs in enumerate(validation_sets):
                if isinstance(vs, Dataset):
                    assert db is not None
                    db.add(vs)
                    validation_sets[i] = vs.identifier

        if db is not None:
            db.add(self)

        if distributed is None:
            distributed = s.CFG.cluster.distributed

        if distributed:
            return self.create_fit_job(
                X,
                select=select,
                y=y,
                **kwargs,
            )(db=db, distributed=True, dependencies=dependencies)
        else:
            return self._fit(
                X,
                y=y,
                configuration=configuration,
                data_prefetch=data_prefetch,
                db=db,
                metrics=metrics,
                select=select,
                validation_sets=validation_sets,
                **kwargs,
            )


@wraps(_TrainingConfiguration)
def TrainingConfiguration(identifier: str, **kwargs):
    return _TrainingConfiguration(identifier=identifier, kwargs=kwargs)



from ibis.backends.base import BaseBackend
from pymongo import MongoClient

from superduperdb.backends.filesystem.artifacts import FileSystemArtifactStore
from superduperdb.backends.ibis.data_backend import IbisDataBackend
from superduperdb.backends.mongodb.artifacts import MongoArtifactStore
from superduperdb.backends.mongodb.data_backend import MongoDataBackend
from superduperdb.backends.mongodb.metadata import MongoMetaDataStore
from superduperdb.backends.sqlalchemy.metadata import SQLAlchemyMetadata
from superduperdb.vector_search.in_memory import InMemoryVectorSearcher
from superduperdb.vector_search.lance import LanceVectorSearcher

data_backends = {'mongodb': MongoDataBackend, 'ibis': IbisDataBackend}

artifact_stores = {'mongodb': MongoArtifactStore, 'filesystem': FileSystemArtifactStore}

metadata_stores = {'mongodb': MongoMetaDataStore, 'sqlalchemy': SQLAlchemyMetadata}

vector_searcher_implementations = {
    'lance': LanceVectorSearcher,
    'in_memory': InMemoryVectorSearcher,
}

CONNECTIONS = {
    'pymongo': MongoClient,
    'ibis': BaseBackend,
}



import dataclasses as dc
import inspect
import logging
import typing as t
import uuid

import click
import requests

from superduperdb.backends.base.artifact import ArtifactStore
from superduperdb.backends.base.query import Delete, Insert, Like, Select, Update
from superduperdb.base.datalayer import ExecuteQuery
from superduperdb.base.document import dump_bsons, load_bsons
from superduperdb.base.serializable import Serializable
from superduperdb.components.component import Component
from superduperdb.misc.serialization import serializers


class ClientArtifactStore:
    def __init__(self, get, put, request_id):
        self.request_id = request_id
        self.get = get
        self.put = put

    def load_artifact(self, file_id, serializer, info=None):
        bytes = self.get(self.request_id, file_id)
        if info in inspect.signature(serializers[serializer]).parameters:
            return serializers[serializer].decode(bytes, info=info)
        else:
            return serializers[serializer].decode(bytes)

    def create_artifact(self, bytes):
        d = self.put(self.request_id, bytes).json()
        return d['file_id'], d['sha1']


class Client:
    def __init__(self, uri, requests=requests):
        """
        :param uri: uri of the server
        """
        self.uri = uri
        self.encoders = LoadDict(self, 'encoder')
        self.requests = requests

    def execute(self, query: ExecuteQuery):
        if isinstance(query, Delete):
            return self.delete(query)
        if isinstance(query, Insert):
            return self.insert(query)
        if isinstance(query, Select):
            return self.select(query)
        if isinstance(query, Like):
            return self.like(query)
        if isinstance(query, Update):
            return self.update(query)
        raise TypeError(
            f'Wrong type of {query}; '
            f'Expected object of type {t.Union[Select, Delete, Update, Insert]}; '
            f'Got {type(query)};'
        )

    def select(self, query: Select):
        """
        Send a request to the server to execute a select query and then
        retrieve the data from the server.

        The request is split into three parts:
        1. Send the serialized query to the server for execution
        2. Retrieve the serialized documents from the server
        3. Deserialize the documents

        Each serialized query contains the following information:
        - request_id: id of the request, randomly generated
        - query: serialized query type

        The request for the serialized documents contains the following information:
        - request_id: id of the request, randomly generated
        - file_id: id of the documents, returned by the initial server response

        :param query: query type to be sent to the server

        :return: deserialized documents
        """
        request_id = str(uuid.uuid4())
        response = self._make_get_request(
            'select',
            json={
                'query': query.serialize(),
                'request_id': request_id,
            },
        ).json()
        result = self._get(request_id=request_id, file_id=response['file_id'])
        documents = load_bsons(result, encoders=self.encoders)
        return documents

    def insert(self, query: Insert):
        """
        The request is split into two parts:
        1. Send serialized documents to the server in a single request
        2. Send serialized metadata to the server, including the query type

        Each serialized sequence of documents request contains the following
        information:
        - request_id: id of the request, randomly generated
        - file_id: extra randomly generated id (TODO: redundant?)
        - data: serialized documents

        The metadata request contains the following information:
        - request_id: id of the request, randomly generated
        - query: serialized query type
        - documents: ID for the documents on the server

        :param query: query type and attached documents to be sent to the server
        """
        documents = dump_bsons(query.documents)
        query.documents = []
        serialized = query.serialize()
        file_id = str(uuid.uuid4())
        request_id = str(uuid.uuid4())
        self._put(request_id=request_id, file_id=file_id, data=documents)
        return self._make_post_or_put_request(
            'insert',
            method='POST',
            json={
                'query': serialized,
                'documents': file_id,
                'request_id': request_id,
            },
        )

    def delete(self, query: Delete):
        """
        Send serialized delete query to the server

        :param query: delete query to be sent to the server
        """
        self._make_post_or_put_request(
            'delete', method='POST', json={'query': query.serialize()}
        )
        return 'ok'

    def _get(self, request_id, file_id):
        """
        Retrieve serialized artifact from the server

        :param request_id: id of the request that created the artifact on the server
        :param file_id: id of the artifact

        :return: serialized artifact
        """
        response = self._make_get_request(
            f'artifacts/get/{request_id}/{file_id}',
        )
        return response.content

    def _put(self, request_id, file_id, data):
        """
        Send serialized artifact to the server

        :param request_id: id of the request, randomly generated
        :param file_id: id of the artifact, randomly generated
        :param data: serialized artifact
        """
        response = self._make_post_or_put_request(
            f'artifacts/put/{request_id}/{file_id}',
            method='PUT',
            data=data,
        )
        return response.text

    def add(self, component: Component):
        """
        Serialize component to metadata and artifacts and send to the server.

        The request is split into two parts:
        1. Send serialized artifacts to the server one-by-one
        2. Send serialized metadata to the server

        Each artifact request contains the following information:
        - request_id: id of the request, randomly generated
        - file_id: id of the file, randomly generated
        - data: serialized artifact

        The metadata request contains the following information:
        - request_id: id of the request, randomly generated
        - component: serialized metadata
        - serializers: mapping from file_id to serializer name eg 'dill'

        :param component: component to be serialized and sent to the server
        """
        request_id = str(uuid.uuid4())
        serialized, artifacts = component.serialized
        lookup = {a.sha1: str(uuid.uuid4()) for a in artifacts}
        serializers = {lookup[a.sha1]: a.serializer for a in artifacts}
        d = ArtifactStore.replace_artifacts_with_dict(serialized, lookup)
        for a in artifacts:
            self._put(
                request_id=request_id,
                file_id=lookup[a.sha1],
                data=a.serialize(),  # Do we need to serialize again?
            )
        self._make_post_or_put_request(
            'add',
            method='POST',
            json={
                'component': d,
                'serializers': serializers,
                'request_id': request_id,
            },
        )

    def show(
        self,
        type_id: str,
        identifier: t.Optional[str] = None,
        version: t.Optional[int] = None,
    ):
        """
        Send show request to the server

        :param type_id: type_id of component to show ['encoder', 'model', 'listener',
                       'learning_task', 'training_configuration', 'metric',
                       'vector_index', 'job']
        :param identifier: identifying string to component
        :param version: (optional) numerical version - specify for full metadata

        :return: metadata of componentd already added to the database
        """
        return self._make_get_request(
            'show',
            json={
                'type_id': type_id,
                'identifier': identifier,
                'version': version,
            },
        ).json()

    def remove(
        self,
        type_id: str,
        identifier: str,
        version: t.Optional[int] = None,
        force: bool = False,
    ):
        """
        Send forced remove request to the server

        :param type_id: type_id of component to show ['encoder', 'model', 'listener',
                       'learning_task', 'training_configuration', 'metric',
                       'vector_index', 'job']
        :param identifier: identifying string to component
        :param version: (optional) numerical version - specify for full metadata
        """
        version_str = '' if version is None else f'/{version}'

        if not force and not click.confirm(
            f'You are about to delete {type_id}/{identifier}{version_str}'
            ', are you sure?',
            default=False,
        ):
            logging.info('aborting...')
            return

        self._make_post_or_put_request(
            'remove',
            method='POST',
            json={
                'type_id': type_id,
                'identifier': identifier,
                'version': version,
            },
        )

    def load(self, type_id: str, identifier: str, version: t.Optional[int] = None):
        """
        Load component from the database via the server.

        This endpoint sends a request to the server to load the component from the
        database. It then receives the component metadata as a response. Then it
        retrieves the component from the server, derserialises the component and
        returns it. It generates a random ID value for communicating with the server.

        :param type_id: type_id of component to load ['encoder', 'model', 'listener',
                       'learning_task', 'training_configuration', 'metric',
                       'vector_index', 'job']
        :param identifier: identifying string to component
        :param version: (optional) numerical version - specify for full metadata
        """
        request_id = str(uuid.uuid4())
        d = self._make_get_request(
            'load',
            json={
                'type_id': type_id,
                'identifier': identifier,
                'version': version,
                'request_id': request_id,
            },
        ).json()
        d = ArtifactStore.load_artifacts(
            d, cache={}, getter=lambda x: self._get(request_id=request_id, file_id=x)
        )
        return Serializable.deserialize(d)

    def like(self, query: Like):
        # TODO: Implement server functionality for this endpoint.
        # TODO: Like.like does not exist, and there is no candidate to replace it
        like = query.like.dump_bson()  # type: ignore[attr-defined]
        query.like = None  # type: ignore[attr-defined]
        serialized = query.serialize()
        file_id = str(uuid.uuid4())
        request_id = str(uuid.uuid4())
        self._put(request_id=request_id, file_id=file_id, data=like)
        out = self._make_get_request(
            'like',
            json={
                'query': serialized,
                'like': file_id,
                'request_id': request_id,
            },
        ).json
        results = self._get(request_id=request_id, file_id=out['file_id'])
        return load_bsons(results, encoders=self.encoders)

    def update(self, query: Update):
        """
        Dump a sequence of documents into BSON and send to the server.

        The request is split into two parts:
        1. Send serialized documents to the server in a single request
        2. Send serialized metadata to the server, including the query type

        Each serialized sequence of documents request contains the following
        information:
        - request_id: id of the request, randomly generated
        - file_id: extra randomly generated id (TODO: redundant?)
        - data: serialized documents

        The metadata request contains the following information:
        - request_id: id of the request, randomly generated
        - query: serialized query type
        - documents: ID for the documents on the server

        :param query: query type and attached documents to be sent to the server
        """
        # TODO: Not sure that this even works...
        # TODO: Update.update does not exist, and there is no candidate to replace it
        update = query.update.dump_bson()  # type: ignore[attr-defined]

        file_id = str(uuid.uuid4())
        request_id = str(uuid.uuid4())
        self._put(request_id=request_id, file_id=file_id, data=update)
        query.update = None  # type: ignore[attr-defined]
        query.serialize()
        return self._make_post_or_put_request(
            'update',
            method='POST',
            json={
                'query': query.serialize(),
                'update': file_id,
                'request_id': request_id,
            },
        )

    def _make_get_request(
        self,
        route: str,
        params: t.Optional[t.Dict] = None,
        json: t.Optional[t.Dict] = None,
    ):
        response = self.requests.get(f'{self.uri}/{route}', params=params, json=json)
        print('YYYYY', response)
        if response.status_code != 200:
            raise ServerSideException(
                f'HTTP status {response.status_code} while making request to'
                f' /{route} with params {params} and json {json}:\n'
                f'{response.text}'
            )
        return response

    def _make_post_or_put_request(
        self,
        route: str,
        method: str,
        json: t.Optional[t.Dict] = None,
        data: t.Optional[t.Dict] = None,
    ):
        if method not in ['PUT', 'POST']:
            raise ServerSideException('Only PUT or POST methods are supported')
        fn = getattr(self.requests, method.lower())
        response = fn(
            f'{self.uri}/{route}',
            data=data,
            json=json,
        )
        if response.status_code != 200:
            raise ServerSideException(
                f'Non 200 status while making request to'
                f' /{route} with json {json}:\n'
                f'{response.text}'
            )
        return response


class ServerSideException(ValueError):
    pass


@dc.dataclass
class LoadDict(dict):
    client: Client
    field: str

    def __missing__(self, key: str):
        value = self[key] = self.client.load(self.field, key)
        return value



import typing as t

from dask import distributed


class DaskClient:
    """
    A client for interacting with a Dask cluster. Initialize the DaskClient.

    :param address: The address of the Dask cluster.
    :param serializers: A list of serializers to be used by the client. (optional)
    :param deserializers: A list of deserializers to be used by the client. (optional)
    :param local: Set to True to create a local Dask cluster. (optional)
    :param envs: An environment dict for cluster.
    :param **kwargs: Additional keyword arguments to be passed to the DaskClient.
    """

    def __init__(
        self,
        address: str,
        serializers: t.Optional[t.Sequence[t.Callable]] = None,
        deserializers: t.Optional[t.Sequence[t.Callable]] = None,
        local: bool = False,
        envs: t.Optional[t.Dict[str, t.Any]] = None,
        **kwargs,
    ):
        envs = envs or {}
        self.futures_collection: t.Dict[str, distributed.Future] = {}

        if local:
            cluster = distributed.LocalCluster(env=envs)
            self.client = distributed.Client(cluster, **kwargs)
        else:
            self.client = distributed.Client(
                address=address,
                serializers=serializers,
                deserializers=deserializers,
                **kwargs,
            )

    def submit(self, function: t.Callable, **kwargs) -> distributed.Future:
        """
        Submits a function to the Dask server for execution.

        :param function: The function to be executed.
        :param kwargs: Additional keyword arguments to be passed to the function.
        """
        future = self.client.submit(function, **kwargs)
        self.futures_collection[future.key] = future
        return future

    def submit_and_forget(self, function: t.Callable, **kwargs) -> distributed.Future:
        """
        Submits a function to the Dask server and keep executing the future
        even if it is no longer referenced.

        :param function: The function to be executed.
        :param kwargs: Additional keyword arguments to be passed to the function.
        """
        future = self.submit(function, **kwargs)
        distributed.fire_and_forget(future)
        return future

    def shutdown(self) -> None:
        """
        Shuts down the Dask client.
        """
        self.client.shutdown()

    def wait_all_pending_tasks(self) -> None:
        """
        Waits for all pending tasks to complete.
        """
        futures = list(self.futures_collection.values())
        distributed.wait(futures)

    def get_result(self, identifier: str) -> t.Any:
        """
        Retrieves the result of a previously submitted task.
        Note: This will block until the future is completed.

        :param identifier: The identifier of the submitted task.
        """
        future = self.futures_collection[identifier]
        return self.client.gather(future)


def dask_client(
    uri: str,
    serializers: t.Optional[t.Sequence[t.Callable]] = None,
    deserializers: t.Optional[t.Sequence[t.Callable]] = None,
    envs: t.Optional[t.Dict[str, t.Any]] = None,
    local: bool = False,
    **kwargs,
) -> DaskClient:
    """
    Creates a DaskClient instance.

    :param uri: The address of the Dask cluster.
    :param serializers: A list of serializers to be used by the client. (optional)
    :param deserializers: A list of deserializers to be used by the client. (optional)
    :param envs: An environment dict for cluster.
    :param local: Set to True to create a local Dask cluster. (optional)
    :param **kwargs: Additional keyword arguments to be passed to the DaskClient.
    """
    return DaskClient(
        address=uri,
        serializers=serializers,
        deserializers=deserializers,
        envs=envs,
        local=local,
        **kwargs,
    )



import dataclasses as dc
import math
import random
import typing as t
import warnings
from collections import defaultdict

import click
import networkx
import tqdm
from dask.distributed import Future

import superduperdb as s
from superduperdb.backends.base.backends import vector_searcher_implementations
from superduperdb.base import serializable
from superduperdb.base.document import Document
from superduperdb.base.logger import logging
from superduperdb.components.component import Component
from superduperdb.components.encoder import Encodable, Encoder
from superduperdb.components.model import Model
from superduperdb.jobs.job import ComponentJob, FunctionJob, Job
from superduperdb.jobs.task_workflow import TaskWorkflow
from superduperdb.misc.colors import Colors
from superduperdb.misc.data import ibatch
from superduperdb.misc.download import Downloader, gather_uris
from superduperdb.misc.special_dicts import MongoStyleDict
from superduperdb.vector_search.base import BaseVectorSearcher, VectorItem
from superduperdb.vector_search.update_tasks import copy_vectors, delete_vectors

from ..backends.base.artifact import ArtifactStore
from ..backends.base.cdc import DatabaseChangeDataCapture
from ..backends.base.data_backend import BaseDataBackend
from ..backends.base.metadata import MetaDataStore
from ..backends.base.query import Delete, Insert, Select, Update
from ..misc.download import download_content
from .cursor import SuperDuperCursor
from .exceptions import ComponentInUseError, ComponentInUseWarning

DBResult = t.Any
TaskGraph = t.Any

DeleteResult = DBResult
InsertResult = t.Tuple[DBResult, t.Optional[TaskGraph]]
SelectResult = SuperDuperCursor
UpdateResult = t.Any

ExecuteQuery = t.Union[Select, Delete, Update, Insert]
ExecuteResult = t.Union[SelectResult, DeleteResult, UpdateResult, InsertResult]

ENDPOINTS = 'delete', 'execute', 'insert', 'like', 'select', 'update'


class Datalayer:
    """
    Base database connector for SuperDuperDB
    """

    type_id_to_cache_mapping = {
        'model': 'models',
        'metric': 'metrics',
        'encoder': 'encoders',
        'vector_index': 'vector_indices',
    }

    def __init__(
        self,
        databackend: BaseDataBackend,
        metadata: MetaDataStore,
        artifact_store: ArtifactStore,
        distributed_client=None,
    ):
        """
        :param databackend: databackend object containing connection to Datastore
        :param metadata: metadata object containing connection to Metadatastore
        :param artifact_store: artifact_store object containing connection to
                               Artifactstore
        :param distributed_client:
        """
        self.metrics = LoadDict(self, field='metric')
        self.models = LoadDict(self, field='model')
        self.encoders = LoadDict(self, field='encoder')
        self.vector_indices = LoadDict(self, field='vector_index')
        self.fast_vector_searchers = LoadDict(
            self, callable=self._initialize_vector_searcher
        )

        self.distributed = s.CFG.cluster.distributed
        self.metadata = metadata
        self.artifact_store = artifact_store
        self.databackend = databackend
        self._distributed_client = distributed_client
        self.cdc = DatabaseChangeDataCapture(self)

    def _initialize_vector_searcher(
        self, identifier, searcher_type: t.Optional[str] = None
    ) -> BaseVectorSearcher:
        searcher_type = searcher_type or s.CFG.vector_search
        logging.info(f'loading of vectors of vector-index: {identifier}')
        vi = self.vector_indices[identifier]

        clt = vi.indexing_listener.select.table_or_collection
        if self.cdc.running:
            msg = 'CDC only supported for vector search via lance format'
            assert s.CFG.vector_search == 'lance', msg

        vector_search_cls = vector_searcher_implementations[searcher_type]
        vector_comparison = vector_search_cls(
            identifier=vi.identifier,
            dimensions=vi.dimensions,
            measure=vi.measure,
        )
        assert isinstance(clt.identifier, str), 'clt.identifier must be a string'

        if self.cdc.running:
            # In this case loading has already happened on disk via CDC mechanism
            return vector_comparison

        if vi.indexing_listener.select is None:
            raise ValueError('.select must be set')

        progress = tqdm.tqdm(desc='Loading vectors into vector-table...')
        for record_batch in ibatch(
            self.execute(
                vi.indexing_listener.select.outputs(
                    vi.indexing_listener.key, vi.indexing_listener.model.identifier
                )
            ),
            s.CFG.cluster.backfill_batch_size,
        ):
            items = []
            for record in record_batch:
                key = vi.indexing_listener.key
                if key.startswith('_outputs.'):
                    key = key.split('.')[1]

                id = record[self.databackend.id_field]
                assert not isinstance(vi.indexing_listener.model, str)
                h = record.outputs(key, vi.indexing_listener.model.identifier)
                if isinstance(h, Encodable):
                    h = h.x
                items.append(VectorItem.create(id=str(id), vector=h))

            vector_comparison.add(items)
            progress.update(len(items))
        return vector_comparison

    @property
    def distributed_client(self):
        return self._distributed_client

    def create_output_table(self, *args):
        """
        Create output table for a model, This is valid for sql databases,
        and databases with seperate output table configuration.
        """
        pass

    def drop(self, force: bool = False):
        """
        Drop all data, artifacts and metadata
        """
        if not force and not click.confirm(
            f'{Colors.RED}[!!!WARNING USE WITH CAUTION AS YOU '
            f'WILL LOSE ALL DATA!!!]{Colors.RESET} '
            'Are you sure you want to drop the database? ',
            default=False,
        ):
            logging.warn('Aborting...')

        self.databackend.drop(force=True)
        self.metadata.drop(force=True)
        self.artifact_store.drop(force=True)

    def validate(
        self,
        identifier: str,
        type_id: str,
        validation_set: str,
        metrics: t.Sequence[str],
    ):
        """
        Evaluate quality of component, using `Component.validate`, if implemented.

        :param identifier: identifier of semantic index
        :param type_id: type_id of component
        :param validation_set: validation dataset on which to validate
        :param metrics: metric functions to compute
        """
        # TODO: never called
        component = self.load(type_id, identifier)
        metric_list = [self.load('metric', m) for m in metrics]
        assert isinstance(component, Model)
        return component.validate(
            self,
            validation_set,
            metric_list,  # type: ignore[arg-type]
        )

    def show(
        self,
        type_id: str,
        identifier: t.Optional[str] = None,
        version: t.Optional[int] = None,
    ):
        """
        Show available functionality which has been added using ``self.add``.
        If version is specified, then print full metadata

        :param type_id: type_id of component to show ['encoder', 'model', 'listener',
                       'learning_task', 'training_configuration', 'metric',
                       'vector_index', 'job']
        :param identifier: identifying string to component
        :param version: (optional) numerical version - specify for full metadata
        """
        if identifier is None and version is not None:
            raise ValueError(f'must specify {identifier} to go with {version}')

        if identifier is None:
            return self.metadata.show_components(type_id=type_id)

        if version is None:
            return self.metadata.show_component_versions(
                type_id=type_id, identifier=identifier
            )

        if version == -1:
            return self._get_object_info(type_id=type_id, identifier=identifier)

        return self._get_object_info(
            type_id=type_id, identifier=identifier, version=version
        )

    def _get_context(
        self, model, context_select: t.Optional[Select], context_key: t.Optional[str]
    ):
        assert model.takes_context, 'model does not take context'
        assert context_select is not None
        context = list(self.execute(context_select))
        context = [x.unpack() for x in context]
        if context_key is not None:
            context = [MongoStyleDict(x)[context_key] for x in context]
        return context

    async def apredict(
        self,
        model_name: str,
        input: t.Union[Document, t.Any],
        context_select: t.Optional[Select] = None,
        context_key: t.Optional[str] = None,
        **kwargs,
    ):
        """
        Apply model to input using asyncio.

        :param model: model identifier
        :param input: input to be passed to the model.
                      Must be possible to encode with registered encoders
        :param context_select: select query object to provide context
        :param context_key: key to use to extract context from context_select
        """
        model = self.models[model_name]
        context = None

        if context_select is not None:
            context = self._get_context(model, context_select, context_key)

        out = await model.apredict(
            input.unpack() if isinstance(input, Document) else input,
            one=True,
            context=context,
            **kwargs,
        )
        if model.encoder is not None:
            out = model.encoder(out)
        if context is not None:
            return Document(out), [Document(x) for x in context]
        return Document(out), []

    def predict(
        self,
        model_name: str,
        input: t.Union[Document, t.Any],
        context_select: t.Optional[Select] = None,
        context_key: t.Optional[str] = None,
        **kwargs,
    ) -> t.Tuple[Document, t.List[Document]]:
        """
        Apply model to input.

        :param model: model identifier
        :param input: input to be passed to the model.
                      Must be possible to encode with registered encoders
        :param context_select: select query object to provide context
        :param context_key: key to use to extract context from context_select
        """
        model = self.models[model_name]
        context = None

        if context_select is not None:
            context = self._get_context(model, context_select, context_key)

        out = model.predict(
            input.unpack() if isinstance(input, Document) else input,
            one=True,
            context=context,
            **kwargs,
        )
        if model.encoder is not None:
            out = model.encoder(out)
        if context is not None:
            return Document(out), [Document(x) for x in context]
        return Document(out), []

    def execute(self, query: ExecuteQuery, *args, **kwargs) -> ExecuteResult:
        """
        Execute a query on the db.

        :param query: select, insert, delete, update,
        """
        if isinstance(query, Delete):
            return self.delete(query, *args, **kwargs)
        if isinstance(query, Insert):
            return self.insert(query, *args, **kwargs)
        if isinstance(query, Select):
            return self.select(query, *args, **kwargs)
        if isinstance(query, Update):
            return self.update(query, *args, **kwargs)
        raise TypeError(
            f'Wrong type of {query}; '
            f'Expected object of type {t.Union[Select, Delete, Update, Insert]}; '
            f'Got {type(query)};'
        )

    def delete(self, delete: Delete, refresh: bool = True) -> DeleteResult:
        """
        Delete data.

        :param delete: delete query object
        """
        result = delete.execute(self)
        if refresh and not self.cdc.running:
            return result, self.refresh_after_delete(delete, ids=result)
        return result, None

    def insert(
        self, insert: Insert, refresh: bool = True, encoders: t.Sequence[Encoder] = ()
    ) -> InsertResult:
        """
        Insert data.

        :param insert: insert query object
        """
        for e in encoders:
            self.add(e)
        for r in insert.documents:
            r['_fold'] = 'train'  # type: ignore[assignment]
            if random.random() < s.CFG.fold_probability:
                r['_fold'] = 'valid'  # type: ignore[assignment]
        inserted_ids = insert.execute(self)

        if refresh and self.cdc.running:
            raise Exception('cdc cannot be activated and refresh=True')

        if refresh:
            return inserted_ids, self.refresh_after_update_or_insert(
                insert, ids=inserted_ids, verbose=False
            )
        return inserted_ids, None

    def run(
        self,
        job,
        depends_on: t.Optional[t.Sequence[Future]] = None,
        distributed: t.Optional[bool] = None,
    ):
        """
        Run job. See ``container.job.Job``, ``container.job.FunctionJob``,
        ``container.job.ComponentJob``.

        :param job:
        :param depends_on: t.Sequence of dependencies
        """
        if distributed is None:
            distributed = s.CFG.cluster.distributed
        return job(db=self, dependencies=depends_on, distributed=distributed)

    def select(self, select: Select) -> SelectResult:
        """
        Select data.

        :param select: select query object
        """
        return select.execute(self)

    def refresh_after_delete(
        self,
        query: Delete,
        ids: t.Sequence[str],
        verbose: bool = False,
    ):
        """
        Trigger cleanup jobs after data deletion.

        :param query: Select or Update which reduces scope of computations
        :param ids: ids which reduce scopy of computations
        :param verbose: Toggle to ``True`` to get more output
        """
        task_workflow: TaskWorkflow = self._build_delete_task_workflow(
            query,
            ids=ids,
            verbose=verbose,
        )
        task_workflow.run_jobs(distributed=self.distributed)
        return task_workflow

    def refresh_after_update_or_insert(
        self,
        query: t.Union[Insert, Select, Update],
        ids: t.Sequence[str],
        verbose: bool = False,
    ):
        """
        Trigger computation jobs after data insertion.

        :param query: Select or Update which reduces scope of computations
        :param ids: ids which reduce scopy of computations
        :param verbose: Toggle to ``True`` to get more output
        """
        task_workflow: TaskWorkflow = self._build_task_workflow(
            query.select_table,  # TODO can be replaced by select_using_ids
            ids=ids,
            verbose=verbose,
        )
        task_workflow.run_jobs(distributed=self.distributed)
        return task_workflow

    def update(self, update: Update, refresh: bool = True) -> UpdateResult:
        """
        Update data.

        :param update: update query object
        """
        updated_ids = update.execute(self)

        if refresh and self.cdc.running:
            raise Exception('cdc cannot be activated and refresh=True')
        if refresh:
            return updated_ids, self.refresh_after_update_or_insert(
                query=update, ids=updated_ids, verbose=False
            )
        return updated_ids, None

    def add(
        self,
        object: t.Union[Component, t.Sequence[Component]],
        dependencies: t.Sequence[Job] = (),
    ):
        """
        Add functionality in the form of components. Components are stored in the
        configured artifact store, and linked to the primary db through
        the metadata.

        :param object: Object to be stored
        :param dependencies: list of jobs which should execute before component
                             init begins
        """
        if isinstance(object, (list, tuple)):
            return type(object)(
                self._add(
                    object=component,
                    dependencies=dependencies,
                )
                for component in object
            )
        elif isinstance(object, Component):
            return self._add(object=object, dependencies=dependencies)
        else:
            raise ValueError(
                'object should be a sequence of `Component` or `Component`'
            )

    def remove(
        self,
        type_id: str,
        identifier: str,
        version: t.Optional[int] = None,
        force: bool = False,
    ):
        """
        Remove component (version: optional)

        :param type_id: type_id of component to remove ['encoder', 'model', 'listener',
                        'training_configuration', 'vector_index']
        :param identifier: identifier of component (see `container.base.Component`)
        :param version: [optional] numerical version to remove
        :param force: force skip confirmation (use with caution)
        """
        if version is not None:
            return self._remove_component_version(type_id, identifier, version=version)
        versions = self.metadata.show_component_versions(type_id, identifier)
        versions_in_use = []
        for v in versions:
            if self.metadata.component_version_has_parents(type_id, identifier, v):
                versions_in_use.append(v)

        if versions_in_use:
            component_versions_in_use = []
            for v in versions_in_use:
                unique_id = Component.make_unique_id(type_id, identifier, v)
                component_versions_in_use.append(
                    f"{unique_id} -> "
                    f"{self.metadata.get_component_version_parents(unique_id)}",
                )
            if not force:
                raise ComponentInUseError(
                    f'Component versions: {component_versions_in_use} are in use'
                )
            else:
                warnings.warn(
                    ComponentInUseWarning(
                        f'Component versions: {component_versions_in_use}'
                        ', marking as hidden'
                    )
                )

        if force or click.confirm(
            f'You are about to delete {type_id}/{identifier}, are you sure?',
            default=False,
        ):
            for v in sorted(list(set(versions) - set(versions_in_use))):
                self._remove_component_version(type_id, identifier, v, force=True)

            for v in sorted(versions_in_use):
                self.metadata.hide_component_version(type_id, identifier, v)
        else:
            logging.warn('aborting.')

    def load(
        self,
        type_id: str,
        identifier: str,
        version: t.Optional[int] = None,
        allow_hidden: bool = False,
        info_only: bool = False,
    ) -> t.Union[Component, t.Dict[str, t.Any]]:
        """
        Load component using uniquely identifying information.

        :param type_id: type_id of component to remove
                        ['encoder', 'model', 'listener', ...]
        :param identifier: identifier of component (see `container.base.Component`)
        :param version: [optional] numerical version
        :param allow_hidden: toggle to ``True`` to allow loading of deprecated
                             components
        :param info_only: toggle to ``True`` to return metadata only
        """
        info = self.metadata.get_component(
            type_id=type_id,
            identifier=identifier,
            version=version,
            allow_hidden=allow_hidden,
        )

        if info is None:
            raise Exception(
                f'No such object of type "{type_id}", '
                f'"{identifier}" has been registered.'
            )

        if info_only:
            return info

        def get_children(info):
            return {
                k: v
                for k, v in info['dict'].items()
                if isinstance(v, dict) and serializable.is_component_metadata(v)
            }

        def replace_children(r):
            if isinstance(r, dict):
                children = get_children(r)
                for k, v in children.items():
                    r['dict'][k] = replace_children(
                        self.metadata.get_component(**v, allow_hidden=True)
                    )
            return r

        info = replace_children(info)
        info = self.artifact_store.load(info, lazy=True)

        m = Component.deserialize(info)
        m.on_load(self)

        if cm := self.type_id_to_cache_mapping.get(type_id):
            getattr(self, cm)[m.identifier] = m
        return m

    def _build_delete_task_workflow(
        self,
        query: Delete,
        ids: t.Sequence[str],
        verbose: bool = False,
    ):
        G = TaskWorkflow(self)
        vector_indices = self.show('vector_index')

        if not vector_indices:
            return G

        deleted_table_or_collection = query.table_or_collection.identifier

        for vi in vector_indices:
            vi = self.vector_indices[vi]
            listener_table_or_collection = (
                vi.indexing_listener.select.table_or_collection.identifier
            )

            if deleted_table_or_collection != listener_table_or_collection:
                continue

            if (
                s.CFG.vector_search == 'in_memory'
                and vi.identifier not in self.fast_vector_searchers
            ):
                continue

            G.add_node(
                f'{vi.identifier}.{delete_vectors.__name__}()',
                job=FunctionJob(
                    callable=delete_vectors,
                    args=[],
                    kwargs=dict(
                        vector_index=vi.identifier,
                        ids=ids,
                    ),
                ),
            )

        return G

    def _build_task_workflow(
        self,
        query,
        ids=None,
        dependencies=(),
        verbose: bool = True,
    ) -> TaskWorkflow:
        job_ids: t.Dict[str, t.Any] = defaultdict(lambda: [])
        job_ids.update(dependencies)
        # TODO use these job_ids as dependencies for every job

        G = TaskWorkflow(self)

        # TODO extract this logic from this class
        try:
            G.add_node(
                f'{download_content.__name__}()',
                job=FunctionJob(
                    callable=download_content,
                    args=[],
                    kwargs=dict(
                        ids=ids,
                        query=query.serialize(),
                    ),
                ),
            )
        except Exception as e:
            logging.info(query)
            raise e

        listeners = self.show('listener')
        if not listeners:
            return G
        listener_selects = {}

        for identifier in listeners:
            info = self.metadata.get_component('listener', identifier)
            listener_query = info['dict']['select']

            listener_select = serializable.Serializable.deserialize(listener_query)
            listener_selects.update({identifier: listener_select})
            if (
                listener_select.table_or_collection.identifier
                != query.table_or_collection.identifier
            ):
                continue

            model, _, key = identifier.rpartition('/')
            G.add_node(
                f'{model}.predict({key})',
                job=ComponentJob(
                    component_identifier=model,
                    args=[key],
                    kwargs={
                        'ids': ids,
                        'select': listener_query,
                        **info['dict']['predict_kwargs'],
                    },
                    method_name='predict',
                    type_id='model',
                ),
            )

        for identifier in listeners:
            listener_select = listener_selects[identifier]
            if (
                listener_select.table_or_collection.identifier
                != query.table_or_collection.identifier
            ):
                continue
            model, _, key = identifier.rpartition('/')
            G.add_edge(
                f'{download_content.__name__}()',
                f'{model}.predict({key})',
            )
            deps = self._get_dependencies_for_listener(
                identifier
            )  # TODO remove features as explicit argument to listener
            for dep in deps:
                dep_model, _, dep_key = dep.rpartition('/')
                G.add_edge(
                    f'{dep_model}.predict({dep_key})',
                    f'{model}.predict({key})',
                )

        if s.CFG.self_hosted_vector_search:
            return G

        for identifier in self.show('vector_index'):
            # if a vector-searcher is not loaded, then skip
            # since s.CFG.vector_search == 'in_memory' implies the
            # program is standalone
            if (
                s.CFG.vector_search == 'in_memory'
                and identifier not in self.fast_vector_searchers
            ):
                continue

            vi = self.vector_indices[identifier]

            G.add_node(
                f'{identifier}.{copy_vectors.__name__}',
                FunctionJob(
                    callable=copy_vectors,
                    args=[],
                    kwargs={
                        'vector_index': identifier,
                        'ids': ids,
                        'query': vi.indexing_listener.select.serialize(),
                    },
                ),
            )
            model = vi.indexing_listener.model.identifier
            key = vi.indexing_listener.key
            G.add_edge(
                f'{model}.predict({key})', f'{identifier}.{copy_vectors.__name__}'
            )

        return G

    def _compute_model_outputs(
        self,
        model_info,
        _ids,
        select: Select,
        key='_base',
        features=None,
        model=None,
        predict_kwargs=None,
    ):
        s.log.info('finding documents under filter')
        features = features or {}
        model_identifier = model_info['identifier']
        if features is None:
            features = {}
        documents = list(self.execute(select.select_using_ids(_ids)))
        s.log.info('done.')
        documents = [x.unpack() for x in documents]
        if key != '_base' or '_base' in features:
            passed_docs = [r[key] for r in documents]
        else:
            passed_docs = documents
        if model is None:
            model = self.models[model_identifier]
        return model.predict(passed_docs, **(predict_kwargs or {}))

    def _save_artifacts(self, artifact_dictionary: t.Dict):
        raise NotImplementedError  # TODO use in the server code...

    def _add(
        self,
        object: Component,
        dependencies: t.Sequence[Job] = (),
        serialized: t.Optional[t.Dict] = None,
        parent: t.Optional[str] = None,
    ):
        object.on_create(self)
        assert hasattr(object, 'identifier')
        assert hasattr(object, 'version')

        existing_versions = self.show(object.type_id, object.identifier)
        if isinstance(object.version, int) and object.version in existing_versions:
            s.log.warn(f'{object.unique_id} already exists - doing nothing')
            return

        if existing_versions:
            object.version = max(existing_versions) + 1
        else:
            object.version = 0

        if serialized is None:
            serialized, artifacts = object.serialized
            artifact_info = self.artifact_store.save(artifacts)
            serialized = self.artifact_store.replace(serialized, artifact_info)

        else:
            serialized['version'] = object.version
            serialized['dict']['version'] = object.version

        self._create_children(object, serialized)
        self.metadata.create_component(serialized)
        if parent is not None:
            self.metadata.create_parent_child(parent, object.unique_id)
        object.on_load(
            self
        )  # TODO do I really need to call this here? Could be handled by `.on_create`?
        jobs = object.schedule_jobs(self, dependencies=dependencies)
        return jobs

    def _create_children(self, component: Component, serialized: t.Dict):
        for k, child_type_id in component.child_components:
            assert isinstance(k, str)
            child = getattr(component, k)
            if isinstance(child, str):
                serialized_dict = {
                    'type_id': child_type_id,
                    'identifier': child,
                    'version': self.metadata.get_latest_version(child_type_id, child),
                }
            else:
                self._add(
                    child,
                    serialized=serialized['dict'][k],
                    parent=component.unique_id,
                )
                serialized_dict = {
                    'type_id': child.type_id,
                    'identifier': child.identifier,
                    'version': child.version,
                }
            serialized['dict'][k] = serialized_dict

    def _create_plan(self):
        G = networkx.DiGraph()
        for identifier in self.metadata.show_components('listener', active=True):
            G.add_node('listener', job=identifier)
        for identifier in self.metadata.show_components('listener'):
            deps = self._get_dependencies_for_listener(identifier)
            for dep in deps:
                G.add_edge(('listener', dep), ('listener', identifier))
        if not networkx.is_directed_acyclic_graph(G):
            raise ValueError('G is not a directed, acyclic graph')
        return G

    def _delete(self, delete: Delete):
        delete.execute(self)

    def _remove_component_version(
        self,
        type_id: str,
        identifier: str,
        version: int,
        force: bool = False,
    ):
        unique_id = Component.make_unique_id(type_id, identifier, version)
        if self.metadata.component_version_has_parents(type_id, identifier, version):
            parents = self.metadata.get_component_version_parents(unique_id)
            raise Exception(f'{unique_id} is involved in other components: {parents}')

        if force or click.confirm(
            f'You are about to delete {unique_id}, are you sure?',
            default=False,
        ):
            component = self.load(type_id, identifier, version=version)
            info = self.metadata.get_component(type_id, identifier, version=version)
            if hasattr(component, 'cleanup'):
                component.cleanup(self)
            if type_id in self.type_id_to_cache_mapping:
                try:
                    del getattr(self, self.type_id_to_cache_mapping[type_id])[
                        identifier
                    ]
                except KeyError:
                    pass

            if hasattr(component, 'artifact_attributes'):
                for a in component.artifact_attributes:
                    self.artifact_store.delete(info['dict'][a]['file_id'])
            self.metadata.delete_component_version(type_id, identifier, version=version)

    def _download_content(  # TODO: duplicated function
        self,
        query: t.Optional[t.Union[Select, Insert]] = None,
        ids=None,
        documents=None,
        timeout=None,
        raises: bool = True,
        n_download_workers=None,
        headers=None,
        **kwargs,
    ):
        update_db = False

        if documents is not None:
            pass
        elif isinstance(query, Select):
            update_db = True
            if ids is None:
                documents = list(self.select(query))
            else:
                select = query.select_using_ids(ids)
                cursor = self.select(select).raw_cursor
                documents = [Document(x) for x in cursor]
        elif isinstance(query, Insert):
            documents = query.documents
        else:
            raise NotImplementedError

        documents = [x.content for x in documents]
        uris, keys, place_ids = gather_uris(documents)
        s.log.info(f'found {len(uris)} uris')
        if not uris:
            return

        if n_download_workers is None:
            try:
                n_download_workers = self.metadata.get_metadata(
                    key='n_download_workers'
                )
            except TypeError:
                n_download_workers = 0

        if headers is None:
            try:
                headers = self.metadata.get_metadata(key='headers')
            except TypeError:
                headers = 0

        if timeout is None:
            try:
                timeout = self.metadata.get_metadata(key='download_timeout')
            except TypeError:
                timeout = None

        def download_update(key, id, bytes):
            return query.download_update(db=self, key=key, id=id, bytes=bytes)

        downloader = Downloader(
            uris=uris,
            ids=place_ids,
            keys=keys,
            update_one=download_update,
            n_workers=n_download_workers,
            timeout=timeout,
            headers=headers,
            raises=raises,
        )
        downloader.go()
        if update_db:
            return
        for id_, key in zip(place_ids, keys):
            documents[id_] = self.databackend.set_content_bytes(
                documents[id_], key, downloader.results[id_]  # type: ignore[index]
            )
        return documents

    def _get_content_for_filter(self, filter) -> Document:
        if isinstance(filter, dict):
            filter = Document(content=filter)
        if '_id' not in filter.content:
            filter['_id'] = 0
        uris = gather_uris([filter.content])[0]
        if uris:
            output = self._download_content(
                query=None, documents=[filter.content], timeout=None, raises=True
            )[0]
            filter = Document(Document.decode(output, encoders=self.encoders))
        if not filter['_id']:
            del filter.content['_id']
        return filter

    def _get_dependencies_for_listener(self, identifier):
        info = self.metadata.get_component('listener', identifier)
        if info is None:
            return []
        listener_features = info.get('features', {})
        out = []
        for k in listener_features:
            out.append(f'{self.features[k]}/{k}')
        if info['dict']['key'].startswith('_outputs.'):
            _, key, model = info['dict']['key'].split('.')
            out.append(f'{model}/{key}')
        return out

    def _get_file_content(self, r):
        for k in r:
            if isinstance(r[k], dict):
                r[k] = self._get_file_content(r[k])
        return r

    def _get_object_info(self, identifier, type_id, version=None):
        return self.metadata.get_component(type_id, identifier, version=version)

    def _apply_listener(
        self,
        identifier,
        ids: t.Optional[t.Sequence[str]] = None,
        verbose: bool = False,
        max_chunk_size=5000,
        model=None,
        recompute: bool = False,
        listener_info=None,
        **kwargs,
    ) -> t.List:
        # NOTE: this method is never called anywhere except for itself!
        if listener_info is None:
            listener_info = self.metadata.get_component('listener', identifier)

        select = serializable.Serializable.deserialize(listener_info['select'])
        if ids is None:
            ids = select.get_ids(self)
        else:
            ids = select.select_using_ids(ids=ids).get_ids(self)
            assert ids is not None

        ids = [str(id) for id in ids]

        if max_chunk_size is not None:
            for it, i in enumerate(range(0, len(ids), max_chunk_size)):
                s.log.info(
                    'computing chunk '
                    f'({it + 1}/{math.ceil(len(ids) / max_chunk_size)})'
                )
                self._apply_listener(
                    identifier,
                    ids=ids[i : i + max_chunk_size],
                    verbose=verbose,
                    max_chunk_size=None,
                    model=model,
                    recompute=recompute,
                    listener_info=listener_info,
                    distributed=False,
                    **kwargs,
                )
            return []

        model_info = self.metadata.get_component('model', listener_info['model'])
        outputs = self._compute_model_outputs(
            model_info,
            ids,
            select,
            key=listener_info['key'],
            features=listener_info.get('features', {}),
            model=model,
            predict_kwargs=listener_info.get('predict_kwargs', {}),
        )
        type = model_info.get('type')
        if type is not None:
            type = self.encoders[type]
            outputs = [type(x).encode() for x in outputs]

        select.model_update(
            db=self,
            model=listener_info['model'],
            key=listener_info['key'],
            outputs=outputs,
            ids=ids,
        )
        return outputs

    def replace(
        self,
        object: t.Any,
        upsert: bool = False,
    ):
        """
        (Use-with caution!!) Replace a model in artifact store with updated object.
        :param object: object to replace
        :param upsert: toggle to ``True`` to enable even if object doesn't exist yet
        """
        try:
            info = self.metadata.get_component(
                object.type_id, object.identifier, version=object.version
            )
        except FileNotFoundError as e:
            if upsert:
                return self.add(
                    object,
                )
            raise e
        new_info = self.artifact_store.update(object, metadata_info=info)
        self.metadata.replace_object(
            new_info,
            identifier=object.identifier,
            type_id='model',
            version=object.version,
        )

    def _select_nearest(
        self,
        like: Document,
        vector_index: str,
        ids: t.Optional[t.Sequence[str]] = None,
        outputs: t.Optional[Document] = None,
        n: int = 100,
    ) -> t.Tuple[t.List[str], t.List[float]]:
        like = self._get_content_for_filter(like)
        vi = self.vector_indices[vector_index]

        if outputs is None:
            outs = {}
        else:
            outs = outputs.encode()
            if not isinstance(outs, dict):
                raise TypeError(f'Expected dict, got {type(outputs)}')
        return vi.get_nearest(like, db=self, ids=ids, n=n, outputs=outs)


@dc.dataclass
class LoadDict(dict):
    database: Datalayer
    field: t.Optional[str] = None
    callable: t.Optional[t.Callable] = None

    def __missing__(self, key: str):
        if self.field is not None:
            value = self[key] = self.database.load(self.field, key)
        else:
            msg = f'callable is ``None`` for {key}'
            assert self.callable is not None, msg
            value = self[key] = self.callable(key)
        return value



from warnings import warn

import ibis
import pandas
from ibis.backends.base import BaseBackend

from superduperdb.backends.base.data_backend import BaseDataBackend
from superduperdb.backends.ibis.field_types import FieldType, dtype
from superduperdb.backends.ibis.query import IbisTable
from superduperdb.components.model import Model
from superduperdb.components.schema import Schema


class IbisDataBackend(BaseDataBackend):
    def __init__(self, conn: BaseBackend, name: str, in_memory: bool = False):
        super().__init__(conn=conn, name=name)
        self.in_memory = in_memory

    def build_artifact_store(self):
        raise NotImplementedError

    def build_metadata(self):
        raise NotImplementedError

    def create_ibis_table(self, identifier: str, schema: Schema):
        self.conn.create_table(identifier, schema=schema)

    def insert(self, table_name, raw_documents):
        if not self.in_memory:
            self.conn.insert(table_name, raw_documents)
        else:
            self.conn.create_table(table_name, pandas.DataFrame(raw_documents))

    def create_model_table_or_collection(self, model: Model):
        msg = (
            "Model must have an encoder to create with the"
            f" {type(self).__name__} backend."
        )
        assert model.encoder is not None, msg
        if isinstance(model.encoder, FieldType):
            output_type = dtype(model.encoder.identifier)
        else:
            output_type = model.encoder
        fields = {
            'output_id': dtype('int32'),
            'input_id': dtype('int32'),
            'query_id': dtype('string'),
            'output': output_type,
            'key': dtype('string'),
        }
        return IbisTable(
            identifier=f'_outputs/{model.identifier}',
            schema=Schema(identifier=f'_schema/{model.identifier}', fields=fields),
        )

    def create_table_and_schema(self, identifier: str, mapping: dict):
        """
        Create a schema in the data-backend.
        """
        try:
            t = self.conn.create_table(identifier, schema=ibis.schema(mapping))
        except Exception as e:
            if 'exists' in str(e):
                warn("Table already exists, skipping...")
                t = self.conn.table(identifier)
            else:
                raise e
        return t

    def drop(self, force: bool = False):
        raise NotImplementedError(
            "Dropping tables needs to be done in each DB natively"
        )

    def get_table_or_collection(self, identifier):
        return self.conn.table(identifier)



"""
The classes in this file define the configuration variables for SuperDuperDB,
which means that this file gets imported before alost anything else, and
canot contain any other imports from this project.
"""

import re
import typing as t
from enum import Enum

from .jsonable import Factory, JSONable


class Retry(JSONable):
    """Describes how to retry using the `tenacity` library

    :param stop_after_attempt: The number of attempts to make
    :param wait_max: The maximum time to wait between attempts
    :param wait_min: The minimum time to wait between attempts
    :param wait_multiplier: The multiplier for the wait time between attempts
    """

    stop_after_attempt: int = 2
    wait_max: float = 10.0
    wait_min: float = 4.0
    wait_multiplier: float = 1.0


class Apis(JSONable):
    """A container for API connections

    :param retry: A ``Retry`` object
    """

    retry: Retry = Factory(Retry)


class Cluster(JSONable):
    """Describes a connection to distributed work via Dask

    :param distributed: Whether to use distributed task management via Dask or not
    :param deserializers: A list of deserializers
    :param serializers: A list of serializers
    :param dask_scheduler: The Dask scheduler URI
    :param local: Whether the connection is local
    """

    distributed: bool = False
    deserializers: t.List[str] = Factory(list)
    serializers: t.List[str] = Factory(list)
    dask_scheduler: str = 'tcp://localhost:8786'
    local: bool = True
    backfill_batch_size: int = 100


class LogLevel(str, Enum):
    """Enumerate log severity level"""

    DEBUG = 'DEBUG'
    INFO = 'INFO'
    WARN = 'WARN'


class LogType(str, Enum):
    """Enumerate the standard logs"""

    STDERR = 'STDERR'
    LOGGING = 'LOGGING'


class Logging(JSONable):
    """Describe how we are going to log. This isn't yet used.

    :param level: The log level
    :param type: The log type
    :param kwargs: Any additional keyword arguments
    """

    level: LogLevel = LogLevel.INFO
    type: LogType = LogType.STDERR
    kwargs: dict = Factory(dict)


class Server(JSONable):
    """Configure the SuperDuperDB server connection information

    :param host: The host for the connection
    :param port: The port for the connection
    :param protocol: The protocol for the connection
    """

    host: str = '127.0.0.1'
    port: int = 3223
    protocol: str = 'http'

    @property
    def uri(self) -> str:
        return f'{self.protocol}://{self.host}:{self.port}'


class Downloads(JSONable):
    """
    Configure how downloads are saved in the database
    or to hybrid filestorage (references to filesystem from datastore)

    :param hybrid: Whether hybrid is being used
    :param root: The root for the connection
    """

    hybrid: bool = False
    root: str = 'data/downloads'


class Config(JSONable):
    """The data class containing all configurable superduperdb values

    :param data_backend: The URI for the data backend
    :param vector_search: The configuration for the vector search {'in_memory', 'lance'}
    :param artifact_store: The URI for the artifact store
    :param metadata_store: The URI for the metadata store
    :param cluster: Settings distributed computing and change data capture
    :param apis: Settings for OPENAI and other APIs
    :param logging: Logging
    :param server: Settings for the experimental Rest server
    :param downloads: Settings for downloading files"""

    @property
    def self_hosted_vector_search(self) -> bool:
        return re.split('://|\+', self.data_backend)[0] == self.vector_search

    data_backend: str = 'mongodb://localhost:27017'

    #: The configuration for the vector search
    vector_search: 'str' = 'in_memory'  # "in_memory" / "lance"

    artifact_store: t.Optional[str] = None
    metadata_store: t.Optional[str] = None
    cluster: Cluster = Factory(Cluster)
    apis: Apis = Factory(Apis)
    logging: Logging = Factory(Logging)
    server: Server = Factory(Server)
    downloads: Downloads = Factory(Downloads)

    #: Probability of validation fold
    fold_probability: float = 0.05

    class Config(JSONable.Config):
        protected_namespaces = ()



import dataclasses as dc
import importlib
import inspect
import typing as t

from superduperdb.misc.serialization import asdict


def is_component_metadata(r: dict) -> bool:
    COMPONENT_KEYS = {'type_id', 'identifier', 'version'}
    if COMPONENT_KEYS == set(r):
        return True
    return False


def is_component(r: dict) -> bool:
    COMPONENT_KEYS = {'cls', 'dict', 'module'}
    if COMPONENT_KEYS <= set(r):
        return True
    return False


def _deserialize(r: t.Any, db: None = None) -> t.Any:
    if isinstance(r, list):
        return [_deserialize(i, db=db) for i in r]

    if not isinstance(r, dict):
        return r
    if not is_component(r):
        return {k: _deserialize(v, db=db) for k, v in r.items()}

    module = importlib.import_module(r['module'])
    component_cls = getattr(module, r['cls'])

    kwargs = _deserialize(r['dict'])
    if 'db' in inspect.signature(component_cls.__init__).parameters:
        kwargs.update(db=db)

    return component_cls(**kwargs)


def _serialize(item: t.Any) -> t.Dict[str, t.Any]:
    def unpack(k, v):
        attr = getattr(item, k)
        if isinstance(attr, Serializable):
            return _serialize(attr)

        if isinstance(attr, list):
            for i, sc in enumerate(attr):
                if isinstance(sc, Serializable):
                    v[i] = _serialize(sc)

        if isinstance(attr, dict):
            for key, value in attr.items():
                if isinstance(value, Serializable):
                    v[key] = _serialize(value)
        return v

    d = {k: unpack(k, v) for k, v in item.dict().items()}

    from superduperdb.components.component import Component

    to_add = {}
    if isinstance(item, Component):
        to_add = {
            'type_id': item.type_id,
            'identifier': item.identifier,
            'version': getattr(item, 'version', None),
        }

    return {
        'cls': item.__class__.__name__,
        'dict': d,
        'module': item.__class__.__module__,
        **to_add,
    }


@dc.dataclass
class Serializable:
    """
    Base class for serializable objects. This class is used to serialize and
    deserialize objects to and from JSON + Artifact instances.
    """

    deserialize = staticmethod(_deserialize)
    serialize = _serialize

    def dict(self) -> t.Dict[str, t.Any]:
        return asdict(self)



from superduperdb import CFG


def maybe_login_required(auth, service):
    """
    Require login depending on the contents of the config file.:w

    :param auth: basic auth instance
    :param service: collection of the service on question
    """

    def decorator(f):
        if getattr(CFG, service).username:
            return auth.login_required(f)
        return f

    return decorator



import dataclasses as dc
import typing as t

import anthropic
from anthropic import APIConnectionError, APIError, APIStatusError, APITimeoutError

from superduperdb.components.component import Component
from superduperdb.components.encoder import Encoder
from superduperdb.components.model import PredictMixin
from superduperdb.ext.utils import format_prompt, get_key
from superduperdb.misc.retry import Retry

retry = Retry(
    exception_types=(APIConnectionError, APIError, APIStatusError, APITimeoutError)
)

KEY_NAME = 'ANTHROPIC_API_KEY'


@dc.dataclass
class Anthropic(Component, PredictMixin):
    """Anthropic predictor.

    :param model: The model to use, e.g. ``'claude-2'``.
    :param identifier: The identifier to use, e.g. ``'my-model'``.
    :param version: The version to use, e.g. ``0`` (leave empty)
    :param takes_context: Whether the model takes context into account.
    :param encoder: The encoder identifier.
    :param type_id: A unique name for the class
    :param client_kwargs: Keyword arguments to pass to the client
    """

    model: str
    identifier: str = ''
    version: t.Optional[int] = None
    takes_context: bool = False
    encoder: t.Union[Encoder, str, None] = None
    client_kwargs: t.Dict[str, t.Any] = dc.field(default_factory=dict)

    type_id: t.ClassVar[str] = 'model'

    def __post_init__(self):
        self.identifier = self.identifier or self.model

    @property
    def child_components(self) -> t.Sequence[t.Tuple[str, str]]:
        if self.encoder is not None:
            return [('encoder', 'encoder')]
        return []


@dc.dataclass
class AnthropicCompletions(Anthropic):
    """Cohere completions (chat) predictor.

    :param takes_context: Whether the model takes context into account.
    :param prompt: The prompt to use to seed the response.
    """

    takes_context: bool = True
    prompt: str = ''

    @retry
    def _predict_one(self, X, context: t.Optional[t.List[str]] = None, **kwargs):
        if context is not None:
            X = format_prompt(X, self.prompt, context=context)
        client = anthropic.Anthropic(api_key=get_key(KEY_NAME), **self.client_kwargs)
        resp = client.completions.create(prompt=X, model=self.identifier, **kwargs)
        return resp.completion

    @retry
    async def _apredict_one(self, X, context: t.Optional[t.List[str]] = None, **kwargs):
        if context is not None:
            X = format_prompt(X, self.prompt, context=context)
        client = anthropic.AsyncAnthropic(
            api_key=get_key(KEY_NAME), **self.client_kwargs
        )
        resp = await client.completions.create(
            prompt=X, model=self.identifier, **kwargs
        )
        return resp.completion

    def _predict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return self._predict_one(X, context=context, **kwargs)
        return [self._predict_one(msg) for msg in X]

    async def _apredict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return await self._apredict_one(X, context=context, **kwargs)
        return [await self._apredict_one(msg) for msg in X]



from __future__ import annotations

import dataclasses as dc
import typing as t
from functools import cached_property

import numpy
from overrides import override

from superduperdb.backends.mongodb.query import Select
from superduperdb.base.artifact import Artifact
from superduperdb.base.document import Document
from superduperdb.components.component import Component

if t.TYPE_CHECKING:
    from superduperdb.base.datalayer import Datalayer


@dc.dataclass
class Dataset(Component):
    """A dataset is an immutable collection of documents that used for training

    :param identifier: A unique identifier for the dataset
    :param select: A query to select the documents for the dataset
    :param sample_size: The number of documents to sample from the query
    :param random_seed: The random seed to use for sampling
    :param creation_date: The date the dataset was created
    :param raw_data: The raw data for the dataset
    :param version: The version of the dataset
    """

    identifier: str
    select: t.Optional[Select] = None
    sample_size: t.Optional[int] = None
    random_seed: t.Optional[int] = None
    creation_date: t.Optional[str] = None
    raw_data: t.Optional[t.Union[Artifact, t.Any]] = None
    version: t.Optional[int] = None

    type_id: t.ClassVar[str] = 'dataset'

    @override
    def on_create(self, db: Datalayer) -> None:
        if self.raw_data is None:
            if self.select is None:
                raise ValueError('select cannot be None')
            data = list(db.execute(self.select))
            if self.sample_size is not None and self.sample_size < len(data):
                perm = self.random.permutation(len(data)).tolist()
                data = [data[perm[i]] for i in range(self.sample_size)]
            self.raw_data = Artifact(artifact=[r.encode() for r in data])

    @override
    def on_load(self, db: Datalayer) -> None:
        assert isinstance(self.raw_data, Artifact)
        self.data = [
            Document(Document.decode(r.copy(), encoders=db.encoders))
            for r in self.raw_data.artifact
        ]

    @cached_property
    def random(self):
        return numpy.random.default_rng(seed=self.random_seed)
